#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2012-2013  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
    base class for generating dialogs for docgen-derived reports
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import os

# -------------------------------------------------------------------------
#
# GTK+ modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.config import config
from ._reportdialog import ReportDialog
from ._papermenu import PaperFrame
from ...pluginmanager import GuiPluginManager

PLUGMAN = GuiPluginManager.get_instance()


# -------------------------------------------------------------------------
#
# DocReportDialog class
#
# -------------------------------------------------------------------------
class DocReportDialog(ReportDialog):
    """
    The DocReportDialog base class.  This is a base class for generating
    dialogs for docgen derived reports.
    """

    def __init__(self, dbstate, uistate, option_class, name, trans_name, track=[]):
        """Initialize a dialog to request that the user select options
        for a basic *stand-alone* report."""

        self.style_name = "default"
        self.firstpage_added = False
        self.css = PLUGMAN.process_plugin_data("WEBSTUFF")
        self.dbname = dbstate.db.get_dbname()
        ReportDialog.__init__(
            self, dbstate, uistate, option_class, name, trans_name, track=track
        )

        self.basedocname = None  # keep pylint happy
        self.css_filename = None
        self.doc = None
        self.doc_option_class = None
        self.format = None
        self.html_label = None
        self.paper_label = None

        # Allow for post processing of the format frame, since the
        # show_all task calls events that may reset values

    def init_interface(self):
        ReportDialog.init_interface(self)
        self.doc_type_changed(self.format_menu, preserve_tab=False)

    # ------------------------------------------------------------------------
    #
    # Functions related to selecting/changing the current file format.
    #
    # ------------------------------------------------------------------------
    def make_doc_menu(self, active=None):
        """Build a menu of document types that are appropriate for
        this report.  This menu will be generated based upon the type
        of document (text, draw, graph, etc. - a subclass), whether or
        not the document requires table support, etc."""
        raise NotImplementedError

    def make_document(self):
        """Create a document of the type requested by the user."""
        pstyle = self.paper_frame.get_paper_style()

        if self.doc_options:
            self.doc = self.format(
                self.selected_style, pstyle, self.doc_options, uistate=self.uistate
            )
        else:
            self.doc = self.format(self.selected_style, pstyle, uistate=self.uistate)
        if not self.format_menu.get_active_plugin().get_paper_used():
            # set css filename
            self.doc.set_css_filename(self.css_filename)

        self.options.set_document(self.doc)

    def doc_type_changed(self, obj, preserve_tab=True):
        """This routine is called when the user selects a new file
        format for the report.  It adjusts the various dialog sections
        to reflect the appropriate values for the currently selected
        file format.  For example, an HTML document doesn't need any
        paper size/orientation options, but it does need a css
        file.  Those changes are made here."""
        docgen_plugin = obj.get_active_plugin()
        if docgen_plugin.get_extension():
            self.open_with_app.set_sensitive(True)
        else:
            self.open_with_app.set_sensitive(False)

        # Is this to be a printed report or an electronic report
        # (i.e. a set of web pages)

        old_page = self.notebook.get_current_page()
        if self.firstpage_added:
            self.notebook.remove_page(0)
        if docgen_plugin.get_paper_used():
            self.paper_label = Gtk.Label(label="<b>%s</b>" % _("Paper Options"))
            self.paper_label.set_use_markup(True)
            self.notebook.insert_page(self.paper_frame, self.paper_label, 0)
            self.paper_frame.show_all()
        else:
            self.html_label = Gtk.Label(label="<b>%s</b>" % _("HTML Options"))
            self.html_label.set_use_markup(True)
            self.notebook.insert_page(self.html_grid, self.html_label, 0)
            self.html_grid.show_all()
        if preserve_tab:
            self.notebook.set_current_page(old_page)
        self.firstpage_added = True

        ext_val = docgen_plugin.get_extension()
        if ext_val:
            fname = self.target_fileentry.get_full_path(0)
            (spath, ext) = os.path.splitext(fname)

            fname = spath + "." + ext_val
            self.target_fileentry.set_filename(fname)
            self.target_fileentry.set_sensitive(True)
        else:
            self.target_fileentry.set_filename("")
            self.target_fileentry.set_sensitive(False)

        # Does this report format use styles?
        if self.style_button:
            self.style_button.set_sensitive(docgen_plugin.get_style_support())
            self.style_menu.set_sensitive(docgen_plugin.get_style_support())

        self.basedocname = docgen_plugin.get_basedocname()
        self.doc_option_class = docgen_plugin.get_doc_option_class()
        self.setup_doc_options_frame()
        self.show()

    def setup_format_frame(self):
        """Set up the format frame of the dialog.  This function
        relies on the make_doc_menu() function to do all the hard
        work."""

        self.make_doc_menu(self.options.handler.get_format_name())
        self.format_menu.connect("changed", self.doc_type_changed)
        label = Gtk.Label(label=_("%s:") % _("Output Format"))
        label.set_halign(Gtk.Align.START)
        self.grid.attach(label, 1, self.row, 1, 1)
        self.format_menu.set_hexpand(True)
        self.grid.attach(self.format_menu, 2, self.row, 2, 1)
        self.row += 1

        self.open_with_app = Gtk.CheckButton(label=_("Open with default viewer"))
        self.open_with_app.set_active(config.get("interface.open-with-default-viewer"))
        self.grid.attach(self.open_with_app, 2, self.row, 2, 1)
        self.row += 1

        ext = self.format_menu.get_active_plugin().get_extension()
        if ext is None:
            ext = ""
        else:
            spath = self.get_default_directory()
            if self.options.get_output():
                base = os.path.basename(self.options.get_output())
            else:
                if self.dbname is None:
                    default_name = self.raw_name
                else:
                    default_name = self.dbname + "_" + self.raw_name
                base = "%s.%s" % (default_name, ext)
            spath = os.path.normpath(os.path.join(spath, base))
            self.target_fileentry.set_filename(spath)

    def setup_report_options_frame(self):
        """Set up the html/paper frame of the dialog"""
        self.paper_frame = PaperFrame(
            self.options.handler.get_paper_metric(),
            self.options.handler.get_paper_name(),
            self.options.handler.get_orientation(),
            self.options.handler.get_margins(),
            self.options.handler.get_custom_paper_size(),
        )
        self.setup_html_frame()
        ReportDialog.setup_report_options_frame(self)

    def setup_html_frame(self):
        """Set up the html frame of the dialog.  This sole purpose of
        this function is to grab a pointer for later use in the parse
        html frame function."""

        self.html_grid = Gtk.Grid()
        self.html_grid.set_column_spacing(12)
        self.html_grid.set_row_spacing(6)
        self.html_grid.set_border_width(6)

        label = Gtk.Label(label=_("%s:") % _("CSS file"))
        label.set_halign(Gtk.Align.START)
        self.html_grid.attach(label, 1, 1, 1, 1)

        self.css_combo = Gtk.ComboBoxText()
        self.css_combo.set_hexpand(True)

        self.css_filename = self.options.handler.get_css_filename()
        active_index = 0
        index = 0
        for name, _id in sorted(
            [(self.css[key]["translation"], self.css[key]["id"]) for key in self.css]
        ):
            if self.css[_id]["user"]:
                self.css_combo.append_text(self.css[_id]["translation"])
                # Associate this index number with CSS too:
                self.css[index] = self.css[_id]
                if self.css_filename == self.css[_id]["filename"]:
                    active_index = index
                index += 1

        self.html_grid.attach(self.css_combo, 2, 1, 1, 1)
        self.css_combo.set_active(active_index)

    def parse_format_frame(self):
        """Parse the format frame of the dialog.  Save the user
        selected output format for later use."""
        docgen_plugin = self.format_menu.get_active_plugin()
        self.basedocname = docgen_plugin.get_basedocname()
        self.format = docgen_plugin.get_basedoc()
        format_name = docgen_plugin.get_extension()
        self.options.handler.set_format_name(format_name)

    def parse_html_frame(self):
        """Parse the html frame of the dialog.  Save the user selected
        html template name for later use.  Note that this routine
        retrieves a value whether or not the file entry box is
        displayed on the screen.  The subclass will know whether this
        entry was enabled.  This is for simplicity of programming."""

        active = self.css_combo.get_active()
        if active == -1:  # legal for "no active item" (see 7585, 8189, 9461)
            active = self.style_name
        if self.css:
            self.css_filename = self.css[active]["filename"]
        self.options.handler.set_css_filename(self.css_filename)

    def on_ok_clicked(self, obj):
        """The user is satisfied with the dialog choices.  Validate
        the output file name before doing anything else.  If there is
        a file name, gather the options and create the report."""

        # Is there a filename?  This should also test file permissions, etc.
        if not self.parse_target_frame():
            self.window.run()

        # Preparation
        self.parse_format_frame()
        self.parse_style_frame()
        self.parse_html_frame()

        self.options.handler.set_paper_metric(self.paper_frame.get_paper_metric())
        self.options.handler.set_paper_name(self.paper_frame.get_paper_name())
        self.options.handler.set_orientation(self.paper_frame.get_orientation())
        self.options.handler.set_margins(self.paper_frame.get_paper_margins())
        self.options.handler.set_custom_paper_size(
            self.paper_frame.get_custom_paper_size()
        )

        self.parse_user_options()

        # Create the output document.
        self.make_document()

        self.parse_doc_options()

        # Save options
        self.options.handler.save_options()
        config.set(
            "interface.open-with-default-viewer", self.open_with_app.get_active()
        )



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A utility to make a best guess if a person is alive.  This is used to provide
privacy in reports and exports.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".gen.utils.alive")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..display.name import displayer as name_displayer
from ..lib.date import Date, Today
from ..errors import DatabaseError
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# -------------------------------------------------------------------------
#
# Constants from config .ini keys
#
# -------------------------------------------------------------------------
# cache values; use refresh_constants() if they change
try:
    from ..config import config

    _MAX_AGE_PROB_ALIVE = config.get("behavior.max-age-prob-alive")
    _MAX_SIB_AGE_DIFF = config.get("behavior.max-sib-age-diff")
    _AVG_GENERATION_GAP = config.get("behavior.avg-generation-gap")
except ImportError:
    # Utils used as module not part of GRAMPS
    _MAX_AGE_PROB_ALIVE = 110
    _MAX_SIB_AGE_DIFF = 20
    _AVG_GENERATION_GAP = 20


# -------------------------------------------------------------------------
#
# ProbablyAlive class
#
# -------------------------------------------------------------------------
class ProbablyAlive:
    """
    An object to hold the parameters for considering someone alive.
    """

    def __init__(
        self,
        db,
        max_sib_age_diff=None,
        max_age_prob_alive=None,
        avg_generation_gap=None,
    ):
        self.db = db
        if max_sib_age_diff is None:
            max_sib_age_diff = _MAX_SIB_AGE_DIFF
        if max_age_prob_alive is None:
            max_age_prob_alive = _MAX_AGE_PROB_ALIVE
        if avg_generation_gap is None:
            avg_generation_gap = _AVG_GENERATION_GAP
        self.MAX_SIB_AGE_DIFF = max_sib_age_diff
        self.MAX_AGE_PROB_ALIVE = max_age_prob_alive
        self.AVG_GENERATION_GAP = avg_generation_gap
        self.pset = set()

    def probably_alive_range(self, person, is_spouse=False):
        # FIXME: some of these computed dates need to be a span. For
        #        example, if a person could be born +/- 20 yrs around
        #        a date then it should be a span, and yr_offset should
        #        deal with it as well ("between 1920 and 1930" + 10 =
        #        "between 1930 and 1940")
        if person is None:
            return (None, None, "", None)
        self.pset = set()
        birth_ref = person.get_birth_ref()
        death_ref = person.get_death_ref()
        death_date = None
        birth_date = None
        explain = ""
        # If the recorded death year is before current year then
        # things are simple.
        if death_ref and death_ref.get_role().is_primary():
            if death_ref:
                death = self.db.get_event_from_handle(death_ref.ref)
                if death:
                    death_date = death.get_date_object()

        # Look for Cause Of Death, Burial or Cremation events.
        # These are fairly good indications that someone's not alive.
        if not death_date:
            for ev_ref in person.get_primary_event_ref_list():
                if ev_ref:
                    ev = self.db.get_event_from_handle(ev_ref.ref)
                    if ev and ev.type.is_death_fallback():
                        death_date = ev.get_date_object()
                        if not death_date.is_valid():
                            death_date = Today()  # before today
                            death_date.set_modifier(Date.MOD_BEFORE)

        # If they were born within X years before current year then
        # assume they are alive (we already know they are not dead).
        if not birth_date:
            if birth_ref and birth_ref.get_role().is_primary():
                birth = self.db.get_event_from_handle(birth_ref.ref)
                if birth and birth.get_date_object().get_start_date() != Date.EMPTY:
                    birth_date = birth.get_date_object()

        # Look for Baptism, etc events.
        # These are fairly good indications that someone's birth.
        if not birth_date:
            for ev_ref in person.get_primary_event_ref_list():
                ev = self.db.get_event_from_handle(ev_ref.ref)
                if ev and ev.type.is_birth_fallback():
                    birth_date = ev.get_date_object()

        if not birth_date and death_date:
            # person died more than MAX after current year
            if death_date.is_valid():
                birth_date = death_date.copy_offset_ymd(year=-self.MAX_AGE_PROB_ALIVE)
            else:
                birth_date = death_date
            explain = _("death date")

        if not death_date and birth_date:
            # person died more than MAX after current year
            death_date = birth_date.copy_offset_ymd(year=self.MAX_AGE_PROB_ALIVE)
            explain = _("birth date")

        if death_date and birth_date:
            return (birth_date, death_date, explain, person)  # direct self evidence

        # Neither birth nor death events are available. Try looking
        # at siblings. If a sibling was born more than X years past,
        # or more than Z future, then probably this person is
        # not alive. If the sibling died more than X years
        # past, or more than X years future, then probably not alive.

        family_list = person.get_parent_family_handle_list()
        for family_handle in family_list:
            family = self.db.get_family_from_handle(family_handle)
            if family is None:
                continue
            for child_ref in family.get_child_ref_list():
                child_handle = child_ref.ref
                child = self.db.get_person_from_handle(child_handle)
                if child is None:
                    continue
                # Go through once looking for direct evidence:
                for ev_ref in child.get_primary_event_ref_list():
                    ev = self.db.get_event_from_handle(ev_ref.ref)
                    if ev and ev.type.is_birth():
                        dobj = ev.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            # if sibling birth date too far away, then not alive:
                            year = dobj.get_year()
                            if year != 0:
                                # sibling birth date
                                return (
                                    Date().copy_ymd(year - self.MAX_SIB_AGE_DIFF),
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("sibling birth date"),
                                    child,
                                )
                    elif ev and ev.type.is_death():
                        dobj = ev.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            # if sibling death date too far away, then not alive:
                            year = dobj.get_year()
                            if year != 0:
                                # sibling death date
                                return (
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        - self.MAX_AGE_PROB_ALIVE
                                    ),
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        - self.MAX_AGE_PROB_ALIVE
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("sibling death date"),
                                    child,
                                )
                # Go through again looking for fallback:
                for ev_ref in child.get_primary_event_ref_list():
                    ev = self.db.get_event_from_handle(ev_ref.ref)
                    if ev and ev.type.is_birth_fallback():
                        dobj = ev.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            # if sibling birth date too far away, then not alive:
                            year = dobj.get_year()
                            if year != 0:
                                # sibling birth date
                                return (
                                    Date().copy_ymd(year - self.MAX_SIB_AGE_DIFF),
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("sibling birth-related date"),
                                    child,
                                )
                    elif ev and ev.type.is_death_fallback():
                        dobj = ev.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            # if sibling death date too far away, then not alive:
                            year = dobj.get_year()
                            if year != 0:
                                # sibling death date
                                return (
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        - self.MAX_AGE_PROB_ALIVE
                                    ),
                                    Date().copy_ymd(
                                        year
                                        - self.MAX_SIB_AGE_DIFF
                                        - self.MAX_AGE_PROB_ALIVE
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("sibling death-related date"),
                                    child,
                                )

        if not is_spouse:  # if you are not in recursion, let's recurse:
            for family_handle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    mother_handle = family.get_mother_handle()
                    father_handle = family.get_father_handle()
                    if mother_handle == person.handle and father_handle:
                        father = self.db.get_person_from_handle(father_handle)
                        date1, date2, explain, other = self.probably_alive_range(
                            father, is_spouse=True
                        )
                        if date1 and date1.get_year() != 0:
                            return (
                                Date().copy_ymd(
                                    date1.get_year() - self.AVG_GENERATION_GAP
                                ),
                                Date().copy_ymd(
                                    date1.get_year()
                                    - self.AVG_GENERATION_GAP
                                    + self.MAX_AGE_PROB_ALIVE
                                ),
                                _("a spouse's birth-related date, ") + explain,
                                other,
                            )
                        elif date2 and date2.get_year() != 0:
                            return (
                                Date().copy_ymd(
                                    date2.get_year()
                                    + self.AVG_GENERATION_GAP
                                    - self.MAX_AGE_PROB_ALIVE
                                ),
                                Date().copy_ymd(
                                    date2.get_year() + self.AVG_GENERATION_GAP
                                ),
                                _("a spouse's death-related date, ") + explain,
                                other,
                            )
                    elif father_handle == person.handle and mother_handle:
                        mother = self.db.get_person_from_handle(mother_handle)
                        date1, date2, explain, other = self.probably_alive_range(
                            mother, is_spouse=True
                        )
                        if date1 and date1.get_year() != 0:
                            return (
                                Date().copy_ymd(
                                    date1.get_year() - self.AVG_GENERATION_GAP
                                ),
                                Date().copy_ymd(
                                    date1.get_year()
                                    - self.AVG_GENERATION_GAP
                                    + self.MAX_AGE_PROB_ALIVE
                                ),
                                _("a spouse's birth-related date, ") + explain,
                                other,
                            )
                        elif date2 and date2.get_year() != 0:
                            return (
                                Date().copy_ymd(
                                    date2.get_year()
                                    + self.AVG_GENERATION_GAP
                                    - self.MAX_AGE_PROB_ALIVE
                                ),
                                Date().copy_ymd(
                                    date2.get_year() + self.AVG_GENERATION_GAP
                                ),
                                _("a spouse's death-related date, ") + explain,
                                other,
                            )
                    # Let's check the family events and see if we find something
                    for ref in family.get_event_ref_list():
                        if ref:
                            event = self.db.get_event_from_handle(ref.ref)
                            if event:
                                date = event.get_date_object()
                                year = date.get_year()
                                if year != 0:
                                    other = None
                                    if person.handle == mother_handle and father_handle:
                                        other = self.db.get_person_from_handle(
                                            father_handle
                                        )
                                    elif (
                                        person.handle == father_handle and mother_handle
                                    ):
                                        other = self.db.get_person_from_handle(
                                            mother_handle
                                        )
                                    return (
                                        Date().copy_ymd(year - self.AVG_GENERATION_GAP),
                                        Date().copy_ymd(
                                            year
                                            - self.AVG_GENERATION_GAP
                                            + self.MAX_AGE_PROB_ALIVE
                                        ),
                                        _("event with spouse"),
                                        other,
                                    )

        # Try looking for descendants that were born more than a lifespan
        # ago.

        def descendants_too_old(person, years):
            if person.handle in self.pset:
                return (None, None, "", None)
            self.pset.add(person.handle)
            for family_handle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                if not family:
                    # can happen with LivingProxyDb(PrivateProxyDb(db))
                    continue
                for child_ref in family.get_child_ref_list():
                    child_handle = child_ref.ref
                    child = self.db.get_person_from_handle(child_handle)
                    child_birth_ref = child.get_birth_ref()
                    if child_birth_ref:
                        child_birth = self.db.get_event_from_handle(child_birth_ref.ref)
                        dobj = child_birth.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            d = Date(dobj)
                            val = d.get_start_date()
                            val = d.get_year() - years
                            d.set_year(val)
                            return (
                                d,
                                d.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE),
                                _("descendant birth date"),
                                child,
                            )
                    child_death_ref = child.get_death_ref()
                    if child_death_ref:
                        child_death = self.db.get_event_from_handle(child_death_ref.ref)
                        dobj = child_death.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            return (
                                dobj.copy_offset_ymd(-self.AVG_GENERATION_GAP),
                                dobj.copy_offset_ymd(
                                    -self.AVG_GENERATION_GAP + self.MAX_AGE_PROB_ALIVE
                                ),
                                _("descendant death date"),
                                child,
                            )
                    date1, date2, explain, other = descendants_too_old(
                        child, years + self.AVG_GENERATION_GAP
                    )
                    if date1 and date2:
                        return date1, date2, explain, other
                    # Check fallback data:
                    for ev_ref in child.get_primary_event_ref_list():
                        ev = self.db.get_event_from_handle(ev_ref.ref)
                        if ev and ev.type.is_birth_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                d = Date(dobj)
                                val = d.get_start_date()
                                val = d.get_year() - years
                                d.set_year(val)
                                return (
                                    d,
                                    d.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE),
                                    _("descendant birth-related date"),
                                    child,
                                )

                        elif ev and ev.type.is_death_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                return (
                                    dobj.copy_offset_ymd(-self.AVG_GENERATION_GAP),
                                    dobj.copy_offset_ymd(
                                        -self.AVG_GENERATION_GAP
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("descendant death-related date"),
                                    child,
                                )

            return (None, None, "", None)

        # If there are descendants that are too old for the person to have
        # been alive in the current year then they must be dead.

        date1, date2, explain, other = None, None, "", None
        try:
            date1, date2, explain, other = descendants_too_old(
                person, self.AVG_GENERATION_GAP
            )
        except RuntimeError:
            raise DatabaseError(
                _("Database error: loop in %s's descendants")
                % name_displayer.display(person)
            )

        if date1 and date2:
            return (date1, date2, explain, other)

        def ancestors_too_old(person, year):
            if person.handle in self.pset:
                return (None, None, "", None)
            self.pset.add(person.handle)
            LOG.debug(
                "ancestors_too_old('%s', %s)", name_displayer.display(person), year
            )
            family_handle = person.get_main_parents_family_handle()
            if family_handle:
                family = self.db.get_family_from_handle(family_handle)
                if not family:
                    # can happen with LivingProxyDb(PrivateProxyDb(db))
                    return (None, None, "", None)
                father_handle = family.get_father_handle()
                if father_handle:
                    father = self.db.get_person_from_handle(father_handle)
                    father_birth_ref = father.get_birth_ref()
                    if father_birth_ref and father_birth_ref.get_role().is_primary():
                        father_birth = self.db.get_event_from_handle(
                            father_birth_ref.ref
                        )
                        dobj = father_birth.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            return (
                                dobj.copy_offset_ymd(-year),
                                dobj.copy_offset_ymd(-year + self.MAX_AGE_PROB_ALIVE),
                                _("ancestor birth date"),
                                father,
                            )
                    father_death_ref = father.get_death_ref()
                    if father_death_ref and father_death_ref.get_role().is_primary():
                        father_death = self.db.get_event_from_handle(
                            father_death_ref.ref
                        )
                        dobj = father_death.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            return (
                                dobj.copy_offset_ymd(-year - self.MAX_AGE_PROB_ALIVE),
                                dobj.copy_offset_ymd(
                                    -year
                                    - self.MAX_AGE_PROB_ALIVE
                                    + self.MAX_AGE_PROB_ALIVE
                                ),
                                _("ancestor death date"),
                                father,
                            )

                    # Check fallback data:
                    for ev_ref in father.get_primary_event_ref_list():
                        ev = self.db.get_event_from_handle(ev_ref.ref)
                        if ev and ev.type.is_birth_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                return (
                                    dobj.copy_offset_ymd(-year),
                                    dobj.copy_offset_ymd(
                                        -year + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("ancestor birth-related date"),
                                    father,
                                )

                        elif ev and ev.type.is_death_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                return (
                                    dobj.copy_offset_ymd(
                                        -year - self.MAX_AGE_PROB_ALIVE
                                    ),
                                    dobj.copy_offset_ymd(
                                        -year
                                        - self.MAX_AGE_PROB_ALIVE
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("ancestor death-related date"),
                                    father,
                                )

                    date1, date2, explain, other = ancestors_too_old(
                        father, year - self.AVG_GENERATION_GAP
                    )
                    if date1 and date2:
                        return date1, date2, explain, other

                mother_handle = family.get_mother_handle()
                if mother_handle:
                    mother = self.db.get_person_from_handle(mother_handle)
                    mother_birth_ref = mother.get_birth_ref()
                    if mother_birth_ref and mother_birth_ref.get_role().is_primary():
                        mother_birth = self.db.get_event_from_handle(
                            mother_birth_ref.ref
                        )
                        dobj = mother_birth.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            return (
                                dobj.copy_offset_ymd(-year),
                                dobj.copy_offset_ymd(-year + self.MAX_AGE_PROB_ALIVE),
                                _("ancestor birth date"),
                                mother,
                            )
                    mother_death_ref = mother.get_death_ref()
                    if mother_death_ref and mother_death_ref.get_role().is_primary():
                        mother_death = self.db.get_event_from_handle(
                            mother_death_ref.ref
                        )
                        dobj = mother_death.get_date_object()
                        if dobj.get_start_date() != Date.EMPTY:
                            return (
                                dobj.copy_offset_ymd(-year - self.MAX_AGE_PROB_ALIVE),
                                dobj.copy_offset_ymd(
                                    -year
                                    - self.MAX_AGE_PROB_ALIVE
                                    + self.MAX_AGE_PROB_ALIVE
                                ),
                                _("ancestor death date"),
                                mother,
                            )

                    # Check fallback data:
                    for ev_ref in mother.get_primary_event_ref_list():
                        ev = self.db.get_event_from_handle(ev_ref.ref)
                        if ev and ev.type.is_birth_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                return (
                                    dobj.copy_offset_ymd(-year),
                                    dobj.copy_offset_ymd(
                                        -year + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("ancestor birth-related date"),
                                    mother,
                                )

                        elif ev and ev.type.is_death_fallback():
                            dobj = ev.get_date_object()
                            if dobj.get_start_date() != Date.EMPTY:
                                return (
                                    dobj.copy_offset_ymd(
                                        -year - self.MAX_AGE_PROB_ALIVE
                                    ),
                                    dobj.copy_offset_ymd(
                                        -year
                                        - self.MAX_AGE_PROB_ALIVE
                                        + self.MAX_AGE_PROB_ALIVE
                                    ),
                                    _("ancestor death-related date"),
                                    mother,
                                )

                    date1, date2, explain, other = ancestors_too_old(
                        mother, year - self.AVG_GENERATION_GAP
                    )
                    if date1 and date2:
                        return (date1, date2, explain, other)

            return (None, None, "", None)

        try:
            # If there are ancestors that would be too old in the current year
            # then assume our person must be dead too.
            date1, date2, explain, other = ancestors_too_old(
                person, -self.AVG_GENERATION_GAP
            )
        except RuntimeError:
            raise DatabaseError(
                _("Database error: loop in %s's ancestors")
                % name_displayer.display(person)
            )
        if date1 and date2:
            return (date1, date2, explain, other)

        # If we can't find any reason to believe that they are dead we
        # must assume they are alive.

        return (None, None, "", None)


# -------------------------------------------------------------------------
#
# probably_alive
#
# -------------------------------------------------------------------------
def probably_alive(
    person,
    db,
    current_date=None,
    limit=0,
    max_sib_age_diff=None,
    max_age_prob_alive=None,
    avg_generation_gap=None,
    return_range=False,
):
    """
    Return true if the person may be alive on current_date.

    This works by a process of elimination. If we can't find a good
    reason to believe that someone is dead then we assume they must
    be alive.

    :param current_date: a date object that is not estimated or modified
                         (defaults to today)
    :param limit: number of years to check beyond death_date
    :param max_sib_age_diff: maximum sibling age difference, in years
    :param max_age_prob_alive: maximum age of a person, in years
    :param avg_generation_gap: average generation gap, in years
    """
    # First, get the real database to use all people
    # for determining alive status:
    birth, death, explain, relative = probably_alive_range(
        person, db, max_sib_age_diff, max_age_prob_alive, avg_generation_gap
    )
    if current_date is None or not current_date.is_valid():
        current_date = Today()
    LOG.debug(
        "%s: b.%s, d.%s - %s",
        name_displayer.display(person),
        birth,
        death,
        explain,
    )
    if not birth or not death:
        # no evidence, must consider alive
        return (True, None, None, _("no evidence"), None) if return_range else True
    # must have dates from here:
    if limit:
        death += limit  # add these years to death
    # Finally, check to see if current_date is between dates
    result = current_date.match(birth, ">=") and current_date.match(death, "<=")
    if return_range:
        return (result, birth, death, explain, relative)
    else:
        return result


def probably_alive_range(
    person, db, max_sib_age_diff=None, max_age_prob_alive=None, avg_generation_gap=None
):
    """
    Computes estimated birth and death dates.
    Returns: (birth_date, death_date, explain_text, related_person)
    """
    # First, find the real database to use all people
    # for determining alive status:
    from ..proxy.proxybase import ProxyDbBase

    basedb = db
    while isinstance(basedb, ProxyDbBase):
        basedb = basedb.db
    # Now, we create a wrapper for doing work:
    pb = ProbablyAlive(basedb, max_sib_age_diff, max_age_prob_alive, avg_generation_gap)
    return pb.probably_alive_range(person)


def update_constants():
    """
    Used to update the constants that are cached in this module.
    """
    from ..config import config

    global _MAX_AGE_PROB_ALIVE, _MAX_SIB_AGE_DIFF, _AVG_GENERATION_GAP
    _MAX_AGE_PROB_ALIVE = config.get("behavior.max-age-prob-alive")
    _MAX_SIB_AGE_DIFF = config.get("behavior.max-sib-age-diff")
    _AVG_GENERATION_GAP = config.get("behavior.avg-generation-gap")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ...pluginmanager import GuiPluginManager
from gramps.gen.plug.report._constants import CATEGORY_TEXT
from ._docreportdialog import DocReportDialog


# -------------------------------------------------------------------------
#
# _TextFormatComboBox
#
# -------------------------------------------------------------------------
class _TextFormatComboBox(Gtk.ComboBox):
    """
    This class is a combo box that allows the selection of a docgen plugin
    from all textdoc plugins.
    """

    def __init__(self, active):
        Gtk.ComboBox.__init__(self)

        pmgr = GuiPluginManager.get_instance()
        self.__textdoc_plugins = []
        for plugin in pmgr.get_docgen_plugins():
            if plugin.get_text_support():
                self.__textdoc_plugins.append(plugin)

        self.store = Gtk.ListStore(GObject.TYPE_STRING)
        self.set_model(self.store)
        cell = Gtk.CellRendererText()
        self.pack_start(cell, True)
        self.add_attribute(cell, "text", 0)

        index = 0
        active_index = 0
        for plugin in self.__textdoc_plugins:
            name = plugin.get_name()
            self.store.append(row=[name])
            if plugin.get_extension() == active:
                active_index = index
            index += 1
        self.set_active(active_index)

    def get_active_plugin(self):
        """
        Get the plugin represented by the currently active selection.
        """
        return self.__textdoc_plugins[self.get_active()]


# -----------------------------------------------------------------------
#
# TextReportDialog
#
# -----------------------------------------------------------------------
class TextReportDialog(DocReportDialog):
    """
    A class of ReportDialog customized for text based reports.
    """

    def __init__(self, dbstate, uistate, options, name, translated_name):
        """
        Initialize a dialog to request that the user select options
        for a basic text report.  See the ReportDialog class for more
        information.
        """
        self.format_menu = None
        self.category = CATEGORY_TEXT
        DocReportDialog.__init__(self, dbstate, uistate, options, name, translated_name)

    def make_doc_menu(self, active=None):
        """
        Build a menu of document types that are appropriate for
        this text report.
        """
        self.format_menu = _TextFormatComboBox(active)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2012       Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
The core of the Gramps plugin system. This module provides capability to load
plugins from specified directories and provide information about the loaded
plugins.

Plugins are divided into several categories. These are: reports, tools,
importers, exporters, quick reports, and document generators.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import os
import sys
import re
import logging
import importlib

LOG = logging.getLogger("._manager")
LOG.progagate = True
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..config import config
from . import PluginRegister, ImportPlugin, ExportPlugin, DocGenPlugin
from ..constfunc import win

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------
_UNAVAILABLE = _("No description was provided")


# -------------------------------------------------------------------------
#
# BasePluginManager
#
# -------------------------------------------------------------------------
class BasePluginManager:
    """unique singleton storage class for a :class:`.PluginManager`."""

    __instance = None

    def get_instance():
        """
        Use this function to get the instance of the :class:`.PluginManager`
        """
        if BasePluginManager.__instance is None:
            BasePluginManager.__instance = 1  # Set to 1 for __init__()
            BasePluginManager.__instance = BasePluginManager()
        return BasePluginManager.__instance

    get_instance = staticmethod(get_instance)

    def __init__(self):
        """This function should only be run once by get_instance()"""
        if BasePluginManager.__instance != 1:
            raise Exception(
                "This class is a singleton. " "Use the get_instance() method"
            )

        self.__import_plugins = []
        self.__export_plugins = []
        self.__docgen_plugins = []

        self.__attempt_list = []
        self.__failmsg_list = []
        self.__external_opt_dict = {}
        self.__success_list = []
        self.__docgen_names = []

        self.__mod2text = {}
        self.__modules = {}

        self.__pgr = PluginRegister.get_instance()
        self.__loaded_plugins = {}
        self.__scanned_dirs = []

    def reg_plugin_dir(
        self, direct, dbstate=None, uistate=None, load_on_reg=False, rescan=False
    ):
        """
        Register plugins in a given directory.
        """
        self.__scanned_dirs.remove(direct)
        self.reg_plugins(direct, dbstate, uistate, load_on_reg, rescan)

    def reg_plugins(
        self, direct, dbstate=None, uistate=None, load_on_reg=False, rescan=False
    ):
        """
        Searches the specified directory, and registers python plugin that
        are being defined in gpr.py files.

        If a relationship calculator for env var LANG is present, it is
        immediately loaded so it is available for all.
        """
        if rescan:
            self.__import_plugins = []
            self.__export_plugins = []
            self.__docgen_plugins = []
            self.__docgen_names = []
            self.__scanned_dirs = []
            self.__pgr._PluginRegister__plugindata = []
            self.__pgr._PluginRegister__id_to_pdata = {}
        # if we've already scanned this directory or if the directory does not
        # exist, we are done.  Should only happen in tests.

        # LOG.warning("\nPlugin manager registration: %s, load_on_reg=%s,"
        #             " been_here=%s, pahte exists:%s", direct, load_on_reg,
        #             direct in self.__scanned_dirs, os.path.isdir(direct))

        if os.path.isdir(direct):
            for dirpath, dirnames, filenames in os.walk(direct, topdown=True):
                for dirname in dirnames[:]:
                    # Skip hidden and system directories:
                    if dirname.startswith(".") or dirname in [
                        "po",
                        "locale",
                        "__pycache__",
                    ]:
                        dirnames.remove(dirname)
                # LOG.warning("Plugin dir scanned: %s", dirpath)
                if dirpath not in self.__scanned_dirs:
                    self.__pgr.scan_dir(dirpath, filenames, uistate=uistate)
                    self.__scanned_dirs.append(dirpath)

        if load_on_reg:
            # Run plugins that request to be loaded on startup and
            # have a load_on_reg callable.
            # first, remove hidden
            plugins_to_load = []
            for plugin in self.__pgr.filter_load_on_reg():
                # LOG.warning("\nFound %s at registration", plugin.id)
                if plugin.id in config.get("plugin.hiddenplugins"):
                    continue
                plugins_to_load.append(plugin)
            # next, sort on dependencies
            # Probably a more efficient method to get dependency graph:
            plugins_sorted = []
            count = 0
            max_count = len(plugins_to_load)
            while plugins_to_load:
                for plugin in plugins_to_load[:]:  # copy of list
                    # LOG.warning("\nDependencies for %s at registration",
                    #             plugin.id)
                    delay = False
                    for depend in plugin.depends_on:
                        if depend not in [p.id for p in plugins_sorted]:
                            delay = True
                            break
                    if delay:
                        pass  # wait till next loop
                    else:
                        if plugin not in plugins_sorted:
                            plugins_sorted.append(plugin)
                        if plugin in plugins_to_load:
                            plugins_to_load.remove(plugin)
                count += 1
                if count > max_count:
                    print("Cannot resolve the following plugin dependencies:")
                    for plugin in plugins_to_load:
                        print(
                            "   Plugin '%s' requires: %s"
                            % (plugin.id, plugin.depends_on)
                        )
                    break
            # now load them:
            for plugin in plugins_sorted:
                # next line shouldn't be necessary, but this gets called a lot
                # of times during Travis test; so avoid multiple copies
                plugin.data = []
                mod = self.load_plugin(plugin)
                if hasattr(mod, "load_on_reg"):
                    # LOG.warning("\nRun %s at registration", plugin.id)
                    try:
                        results = mod.load_on_reg(dbstate, uistate, plugin)
                    except:
                        import traceback

                        traceback.print_exc()
                        print("Plugin '%s' did not run; continuing..." % plugin.name)
                        continue
                    try:
                        iter(results)
                        plugin.data += results
                    except:
                        plugin.data = results
        # Get the addon rules and import them and make them findable
        for plugin in self.__pgr.rule_plugins():
            mod = self.load_plugin(plugin)  # load the addon rule
            # get place in rule heirarchy to put the new rule
            obj_rules = importlib.import_module(
                "gramps.gen.filters.rules." + plugin.namespace.lower()
            )
            # get the new rule class object
            r_class = getattr(mod, plugin.ruleclass)
            # make the new rule findable via import statements
            setattr(obj_rules, plugin.ruleclass, r_class)
            # and add it to the correct fiter editor list
            obj_rules.editor_rule_list.append(r_class)

    def is_loaded(self, pdata_id):
        """
        return True if plugin is already loaded
        """
        if pdata_id in self.__loaded_plugins:
            return True
        return False

    def load_plugin(self, pdata):
        """
        Load a :class:`.PluginData` object. This means import of the python
        module. Plugin directories are added to sys path, so files are found
        """
        if pdata.id in self.__loaded_plugins:
            return self.__loaded_plugins[pdata.id]
        need_reload = False
        filename = pdata.fname
        if filename in self.__modules:
            # filename is loaded already, a different plugin in this module
            _module = self.__modules[filename]
            self.__success_list.append((filename, _module, pdata))
            self.__loaded_plugins[pdata.id] = _module
            self.__mod2text[_module.__name__] += " - " + pdata.description
            return _module
        if filename in self.__attempt_list:
            # new load attempt after a fail, a reload needed
            need_reload = True
            # remove previous fail of the plugins in this file
            dellist = []
            for index, data in enumerate(self.__failmsg_list):
                if data[0] == filename:
                    dellist.append(index)
            dellist.reverse()
            for index in dellist:
                del self.__failmsg_list[index]
        else:
            self.__attempt_list.append(filename)
        try:
            _module = self.import_plugin(pdata)
            if need_reload:
                # For some strange reason second importing of a failed plugin
                # results in success. Then reload reveals the actual error.
                # Looks like a bug in Python.
                _module = self.reload(_module, pdata)
            if _module:
                self.__success_list.append((filename, _module, pdata))
                self.__modules[filename] = _module
                self.__loaded_plugins[pdata.id] = _module
                self.__mod2text[_module.__name__] = pdata.description
            return _module
        except:
            import traceback

            traceback.print_exc()
            self.__failmsg_list.append((filename, sys.exc_info(), pdata))

        return None

    def import_plugin(self, pdata):
        """
        Rather than just __import__(id), this will add the pdata.fpath
        to sys.path first (if needed), import, and then reset path.
        """
        module = None
        if isinstance(pdata, str):
            pdata = self.get_plugin(pdata)
        if not pdata:
            return None
        if pdata.fpath not in sys.path:
            if pdata.mod_name:
                sys.path.insert(0, pdata.fpath)
                try:
                    module = __import__(pdata.mod_name)
                except ValueError as err:
                    # Python3 on Windows  work with unicode in sys.path
                    # but they are mbcs encode for checking validity
                    if win():
                        # we don't want to load Gramps core plugin like this
                        # only 3rd party plugins
                        if "gramps" in pdata.fpath:
                            try:
                                sys.path.insert(0, ".")
                                oldwd = os.getcwd()
                                os.chdir(pdata.fpath)
                                module = __import__(pdata.mod_name)
                                os.chdir(oldwd)
                                sys.path.pop(0)
                            except ValueError as err:
                                LOG.warning(
                                    "Plugin error (from '%s'): %s"
                                    % (pdata.mod_name, err)
                                )
                    else:
                        LOG.warning(
                            "Plugin error (from '%s'): %s" % (pdata.mod_name, err)
                        )
                except ImportError as err:
                    LOG.warning("Plugin error (from '%s'): %s" % (pdata.mod_name, err))
                sys.path.pop(0)
            else:
                print("WARNING: module cannot be loaded")
        else:
            module = __import__(pdata.mod_name)
        return module

    def empty_managed_plugins(self):
        """For some plugins, managed Plugin are used. These are only
        reobtained from the registry if this method is called
        """
        # TODO: do other lists need to be reset here, too?
        self.__import_plugins = []
        self.__export_plugins = []
        self.__docgen_plugins = []
        self.__docgen_names = []

    def reload_plugins(self):
        """Reload previously loaded plugins"""
        pymod = re.compile(r"^(.*)\.py$")

        oldfailmsg = self.__failmsg_list[:]
        self.__failmsg_list = []

        # attempt to reload all plugins that have succeeded in the past
        self.empty_managed_plugins()
        self.__loaded_plugins = {}

        oldmodules = self.__modules
        self.__modules = {}
        dellist = []
        # reload first modules that loaded successfully previously
        for index, plugin in enumerate(self.__success_list):
            filename = plugin[0]
            pdata = plugin[2]
            filename = filename.replace("pyc", "py")
            filename = filename.replace("pyo", "py")
            if filename in self.__modules:
                # module already reloaded, a second plugin in same module
                continue
            try:
                self.reload(plugin[1], pdata)
                self.__modules[filename] = plugin[1]
                self.__loaded_plugins[pdata.id] = plugin[1]
            except:
                dellist.append(index)
                self.__failmsg_list.append((filename, sys.exc_info(), pdata))

        dellist.reverse()
        for index in dellist:
            del self.__success_list[index]

        # Remove previously good plugins that are now bad
        # from the registered lists
        self.__purge_failed()

        # attempt to load the plugins that have failed in the past
        for filename, message, pdata in oldfailmsg:
            self.load_plugin(pdata)

    def reload(self, module, pdata):
        """
        Reloads modules that might not be in the path.
        """
        try:
            spec = importlib.util.find_spec(pdata.mod_name, [pdata.fpath])
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        except:
            if pdata.mod_name in sys.modules:
                del sys.modules[pdata.mod_name]
            module = self.import_plugin(pdata)
        return module

    def get_fail_list(self):
        """Return the list of failed plugins."""
        return self.__failmsg_list

    def get_success_list(self):
        """Return the list of succeeded plugins."""
        return self.__success_list

    def get_plugin(self, id):
        """
        Returns a plugin object from :class:`.PluginRegister` by id.
        """
        return self.__pgr.get_plugin(id)

    def get_reg_reports(self, gui=True):
        """Return list of registered reports
        :param gui: bool indicating if GUI reports or CLI reports must be
        returned
        """
        return self.__pgr.report_plugins(gui)

    def get_reg_tools(self, gui=True):
        """Return list of registered tools
        :aram gui: bool indicating if GUI reports or CLI reports must be
        returned
        """
        return self.__pgr.tool_plugins(gui)

    def get_reg_quick_reports(self):
        """Return list of registered quick reports"""
        return self.__pgr.quickreport_plugins()

    def get_reg_views(self):
        """Return list of registered views"""
        return self.__pgr.view_plugins()

    def get_reg_mapservices(self):
        """Return list of registered mapservices"""
        return self.__pgr.mapservice_plugins()

    def get_reg_bookitems(self):
        """Return list of reports registered as bookitem"""
        return self.__pgr.bookitem_plugins()

    def get_reg_gramplets(self):
        """Return list of non hidden gramplets."""
        return self.__pgr.gramplet_plugins()

    def get_reg_sidebars(self):
        """Return list of registered sidebars."""
        return self.__pgr.sidebar_plugins()

    def get_reg_databases(self):
        """Return list of registered database backends"""
        return self.__pgr.database_plugins()

    def get_reg_thumbnailers(self):
        """Return list of registered thumbnailers."""
        return self.__pgr.thumbnailer_plugins()

    def get_reg_cite(self):
        """Return list of registered cite plugins."""
        return self.__pgr.cite_plugins()

    def get_external_opt_dict(self):
        """Return the dictionary of external options."""
        return self.__external_opt_dict

    def get_module_description(self, module):
        """Given a module name, return the module description."""
        return self.__mod2text.get(module, "")

    def get_reg_importers(self):
        """Return list of registered importers"""
        return self.__pgr.import_plugins()

    def get_reg_exporters(self):
        """Return list of registered exporters"""
        return self.__pgr.export_plugins()

    def get_reg_docgens(self):
        """Return list of registered docgen"""
        return self.__pgr.docgen_plugins()

    def get_reg_general(self, category=None):
        """Return list of registered general libs"""
        return self.__pgr.general_plugins(category)

    def load_plugin_category(self, category):
        """
        Make sure all plugins of a type are loaded.
        """
        for plugin in self.__pgr.general_plugins(category):
            if not self.is_loaded(plugin):
                self.load_plugin(plugin)

    def get_plugin_data(self, category):
        """
        Gets all of the data from general plugins of type category.
        plugin.data may be a single item, an iterable, or a callable.

        >>> PLUGMAN.get_plugin_data('CSS')
        <a list of raw data items>
        """
        retval = []
        data = None
        for plugin in self.__pgr.general_plugins(category):
            data = plugin.data
            try:
                iter(data)
                retval.extend(data)
            except:
                retval.append(data)
        return retval

    def process_plugin_data(self, category):
        """
        Gathers all of the data from general plugins of type category,
        and pass it to a single process function from one of those
        plugins.

        >>> PLUGMAN.process_plugin_data('CSS')
        <a list of processed data items>
        """
        retval = []
        data = None
        process = None
        for plugin in self.__pgr.general_plugins(category):
            if plugin.process is not None:
                mod = self.load_plugin(plugin)
                if hasattr(mod, plugin.process):
                    process = getattr(mod, plugin.process)
            data = plugin.data
            if data:
                try:
                    iter(data)
                    retval.extend(data)
                except:
                    retval.append(data)
        # LOG.warning("Process plugin data=%s, %s, items=%s",
        #             process is not None, category, len(retval))
        if process:
            return process(retval)
        return retval

    def get_import_plugins(self):
        """
        Get the list of import plugins.

        :return: :class:`.ImportPlugin` (a list of ImportPlugin instances)
        """
        ## TODO: would it not be better to remove ImportPlugin and use
        ## only PluginData, loading from module when importfunction needed?
        if self.__import_plugins == []:
            # The module still needs to be imported
            for pdata in self.get_reg_importers():
                if pdata.id in config.get("plugin.hiddenplugins"):
                    continue
                mod = self.load_plugin(pdata)
                if mod:
                    imp = ImportPlugin(
                        name=pdata.name,
                        description=pdata.description,
                        import_function=getattr(mod, pdata.import_function),
                        extension=pdata.extension,
                    )
                    self.__import_plugins.append(imp)

        return self.__import_plugins

    def get_export_plugins(self):
        """
        Get the list of export plugins.

        :return: :class:`.ExportPlugin` (a list of ExportPlugin instances)
        """
        ## TODO: would it not be better to remove ExportPlugin and use
        ## only PluginData, loading from module when export/options needed?
        if self.__export_plugins == []:
            # The modules still need to be imported
            for pdata in self.get_reg_exporters():
                if pdata.id in config.get("plugin.hiddenplugins"):
                    continue
                mod = self.load_plugin(pdata)
                if mod:
                    options = None
                    if pdata.export_options and hasattr(mod, pdata.export_options):
                        options = getattr(mod, pdata.export_options)
                    exp = ExportPlugin(
                        name=pdata.name_accell,
                        description=pdata.description,
                        export_function=getattr(mod, pdata.export_function),
                        extension=pdata.extension,
                        config=(pdata.export_options_title, options),
                    )
                    self.__export_plugins.append(exp)

        return self.__export_plugins

    def get_docgen_plugins(self):
        """
        Get the list of docgen plugins.

        :return: :class:`.DocGenPlugin` (a list of DocGenPlugin instances)
        """
        ## TODO: would it not be better to return list of plugindata, and only
        ##       import those docgen that will then actuallly be needed?
        ##       So, only do import when docgen.get_basedoc() is requested
        if self.__docgen_plugins == []:
            # The modules still need to be imported
            hiddenplugins = config.get("plugin.hiddenplugins")
            for pdata in self.get_reg_docgens():
                if pdata.id in hiddenplugins:
                    continue
                mod = self.load_plugin(pdata)
                if mod:
                    oclass = None
                    if pdata.optionclass:
                        oclass = getattr(mod, pdata.optionclass)
                    dgp = DocGenPlugin(
                        name=pdata.name,
                        description=pdata.description,
                        basedoc=getattr(mod, pdata.docclass),
                        paper=pdata.paper,
                        style=pdata.style,
                        extension=pdata.extension,
                        docoptclass=oclass,
                        basedocname=pdata.docclass,
                    )
                    self.__docgen_plugins.append(dgp)

        return self.__docgen_plugins

    def get_docgen_names(self):
        """
        Get the list of docgen plugin names.

        :return: a list of :class:`.DocGenPlugin` names
        """
        if self.__docgen_names == []:
            hiddenplugins = config.get("plugin.hiddenplugins")
            for pdata in self.get_reg_docgens():
                if pdata.id not in hiddenplugins:
                    self.__docgen_names.append(pdata.docclass)
        return self.__docgen_names

    def register_option(self, option, guioption):
        """
        Register an external option.

        Register a mapping from option to guioption for an option
        that is not native to Gramps but provided by the plugin writer.
        This should typically be called during initialisation of a
        :class:`.ReportOptions` class.

        :param option:      the option class
        :type option:       class that inherits from gen.plug.menu.Option
        :param guioption:   the gui-option class
        :type guioption:    class that inherits from Gtk.Widget.
        """
        self.__external_opt_dict[option] = guioption

    def __purge_failed(self):
        """
        Purge the failed plugins from the corresponding lists.
        """
        failed_module_names = [
            os.path.splitext(os.path.basename(filename))[0]
            for filename, msg, pdata in self.__failmsg_list
        ]

        self.__export_plugins[:] = [
            item
            for item in self.__export_plugins
            if item.get_module_name() not in failed_module_names
        ][:]
        self.__import_plugins[:] = [
            item
            for item in self.__import_plugins
            if item.get_module_name() not in failed_module_names
        ][:]
        self.__docgen_plugins[:] = [
            item
            for item in self.__docgen_plugins
            if item.get_module_name() not in failed_module_names
        ][:]



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2005  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       lcc & Robert Jerome
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# Portuguese version by Duarte Loreto <happyguy_pt@hotmail.com>, 2007.
# Based on the Spanish version by Julio Sanchez <julio.sanchez@gmail.com>
"""
Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------

_level_name_male = [
    "",
    "primeiro",
    "segundo",
    "terceiro",
    "quarto",
    "quinto",
    "sexto",
    "sétimo",
    "oitavo",
    "nono",
    "décimo",
    "décimo-primeiro",
    "décimo-segundo",
    "décimo-terceiro",
    "décimo-quarto",
    "décimo-quinto",
    "décimo-sexto",
    "décimo-sétimo",
    "décimo-oitavo",
    "décimo-nono",
    "vigésimo",
]

# Short forms (in apocope) used before names
_level_name_male_a = [
    "",
    "primeiro",
    "segundo",
    "terceiro",
    "quarto",
    "quinto",
    "sexto",
    "sétimo",
    "oitavo",
    "nono",
    "décimo",
    "décimo-primeiro",
    "décimo-segundo",
    "décimo-terceiro",
    "décimo-quarto",
    "décimo-quinto",
    "décimo-sexto",
    "décimo-sétimo",
    "décimo-oitavo",
    "décimo-nono",
    "vigésimo",
]

_level_name_female = [
    "",
    "primeira",
    "segunda",
    "terceira",
    "quarta",
    "quinta",
    "sexta",
    "sétima",
    "oitava",
    "nona",
    "décima",
    "décima-primeira",
    "décima-segunda",
    "décima-terceira",
    "décima-quarta",
    "décima-quinta",
    "décima-sexta",
    "décima-sétima",
    "décima-oitava",
    "décima-nona",
    "vigésima",
]

_level_name_plural = [
    "",
    "primeiros",
    "segundos",
    "terceiros",
    "quartos",
    "quintos",
    "sextos",
    "sétimos",
    "oitavos",
    "nonos",
    "décimos",
    "décimos-primeiros",
    "décimos-segundos",
    "décimos-terceiros",
    "décimos-quartos",
    "décimos-quintos",
    "décimos-sextos",
    "décimos-sétimos",
    "décimos-oitavos",
    "décimos-nonos",
    "vigésimos",
]

# This plugin tries to be flexible and expect little from the following
# tables.  Ancestors are named from the list for the first generations.
# When this list is not enough, ordinals are used based on the same idea,
# i.e. bisavô is 'segundo avô' and so on, that has been the
# traditional way in Portuguese.  When we run out of ordinals we resort to
# Nº notation, that is sort of understandable if in context.
# There is a specificity for pt_BR where they can use "tataravô" instead
# of "tetravô", being both forms correct for pt_BR but just "tetravô"
# correct for pt. Translation keeps "tetravô".
_parents_level = [
    "",
    "pais",
    "avós",
    "bisavós",
    "trisavós",
    "tetravós",
    "pentavós",
    "hexavós",
    "heptavós",
    "octavós",
]

_father_level = [
    "",
    "pai",
    "avô",
    "bisavô",
    "trisavô",
    "tetravô",
    "pentavô",
    "hexavô",
    "heptavô",
    "octavô",
]

_mother_level = [
    "",
    "mãe",
    "avó",
    "bisavó",
    "trisavó",
    "tetravó",
    "pentavó",
    "hexavó",
    "heptavó",
    "octavó",
]

# Higher-order terms (after "tetravô") are not standard in Portuguese.
# Check http://www.geneall.net/P/forum_msg.php?id=136774 that states
# that although some people may use other greek-prefixed forms for
# higher levels, both pt and pt_BR correct form is to use, after
# "tetravô", the "quinto avô", "sexto avô", etc.

_son_level = [
    "",
    "filho",
    "neto",
    "bisneto",
    "trineto",
    "tetraneto",
    "pentaneto",
    "hexaneto",
    "heptaneto",
    "octaneto",
]

_daughter_level = [
    "",
    "filha",
    "neta",
    "bisneta",
    "trineta",
    "tetraneta",
    "pentaneta",
    "hexaneta",
    "heptaneta",
    "octaneta",
]

_sister_level = [
    "",
    "irmã",
    "tia",
    "tia avó",
    "tia bisavó",
    "tia trisavó",
    "tia tetravó",
    "tia pentavó",
    "tia hexavó",
    "tia heptavó",
    "tia octavó",
]

_brother_level = [
    "",
    "irmão",
    "tio",
    "tio avô",
    "tio bisavô",
    "tio trisavô",
    "tio tetravô",
    "tio pentavô",
    "tio hexavô",
    "tio heptavô",
    "tio octavô",
]

_nephew_level = [
    "",
    "sobrinho",
    "sobrinho neto",
    "sobrinho bisneto",
    "sobrinho trineto",
    "sobrinho tetraneto",
    "sobrinho pentaneto",
    "sobrinho hexaneto",
    "sobrinho heptaneto",
    "sobrinho octaneto",
]

_niece_level = [
    "",
    "sobrinha",
    "sobrinha neta",
    "sobrinha bisneta",
    "sobrinha trineta",
    "sobrinha tetraneta",
    "sobrinha pentaneta",
    "sobrinha hexaneta",
    "sobrinha heptaneta",
    "sobrinha octaneta",
]

# Relatório de Parentesco

_PARENTS_LEVEL = [
    "",
    "pais",
    "avós",
    "bisavós",
    "tetravós",
    "pentavós",
    "hexavós",
    "heptavós",
    "octavós",
]

_CHILDREN_LEVEL = [
    "",
    "filhos",
    "netos",
    "bisnetos",
    "trinetos",
    "tetranetos",
    "pentanetos",
    "hexanetos",
    "heptanetos" "octanetos",
]

_SIBLINGS_LEVEL = [
    "",
    "irmãos e irmãs",
    "tios e tias",
    "tios avôs e tias avós",
    "tios bisavôs e tias bisavós",
    "tios trisavôs e tias trisavós",
    "tios tetravôs e tias tetravós",
    "tios pentavôs e tias pentavós",
    "tios hexavôs e tias hexavós",
    "tios heptavôs e tias heptavós" "tios octavôs e tias octavós",
]

_NEPHEWS_NIECES_LEVEL = [
    "",
    "sobrinhos e sobrinhas",
    "sobrinhos netos e sobrinhas netas",
    "sobrinhos bisnetos e sobrinhas bisnetas",
    "sobrinhos trinetos e sobrinhas trinetas"
    "sobrinhos tetranetos e sobrinhas tetranetas"
    "sobrinhos pentanetos e sobrinhas pentanetas"
    "sobrinhos hexanetos e sobrinhas hexanetas"
    "sobrinhos heptanetos e sobrinhas heptanetas"
    "sobrinhos octanetos e sobrinhas octanetas",
]


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_male_cousin(self, level):
        if level < len(_level_name_male):
            return "%s primo" % (_level_name_male[level])
        else:
            return "%dº primo" % level

    def get_female_cousin(self, level):
        if level < len(_level_name_female):
            return "%s prima" % (_level_name_female[level])
        else:
            return "%dª prima" % level

    def get_distant_uncle(self, level):
        if level < len(_level_name_male):
            return "%s tio" % (_level_name_male[level])
        else:
            return "%dº tio" % level

    def get_distant_aunt(self, level):
        if level < len(_level_name_female):
            return "%s tia" % (_level_name_female[level])
        else:
            return "%dª tia" % level

    def get_distant_nephew(self, level):
        if level < len(_level_name_male):
            return "%s sobrinho" % (_level_name_male[level])
        else:
            return "%dº sobrinho" % level

    def get_distant_niece(self, level):
        if level < len(_level_name_female):
            return "%s sobrinha" % (_level_name_female[level])
        else:
            return "%dª sobrinha" % level

    def get_male_relative(self, level1, level2):
        if level1 < len(_level_name_male_a):
            level1_str = _level_name_male_a[level1]
        else:
            level1_str = "%dº" % level1
        if level2 < len(_level_name_male_a):
            level2_str = _level_name_male_a[level2]
        else:
            level2_str = "%dº" % level2
        level = level1 + level2
        if level < len(_level_name_male_a):
            level_str = _level_name_male_a[level]
        else:
            level_str = "%dº" % level
        return "parente em %s grau (%s com %s)" % (level_str, level1_str, level2_str)

    def get_female_relative(self, level1, level2):
        return self.get_male_relative(level1, level2)

    def get_parents(self, level):
        if level < len(_parents_level):
            return _parents_level[level]
        elif (level - 1) < len(_level_name_plural):
            return "%s avós" % (_level_name_plural[level - 1])
        else:
            return "%dº avós" % (level - 1)

    def get_father(self, level):
        if level < len(_father_level):
            return _father_level[level]
        elif (level - 1) < len(_level_name_male_a):
            return "%s avô" % (_level_name_male_a[level - 1])
        else:
            return "%dº avô" % (level - 1)

    def get_son(self, level):
        if level < len(_son_level):
            return _son_level[level]
        elif (level - 1) < len(_level_name_male_a):
            return "%s neto" % (_level_name_male_a[level - 1])
        else:
            return "%dº neto" % (level - 1)

    def get_mother(self, level):
        if level < len(_mother_level):
            return _mother_level[level]
        elif (level - 1) < len(_level_name_female):
            return "%s avó" % (_level_name_female[level - 1])
        else:
            return "%dª avó" % (level - 1)

    def get_daughter(self, level):
        if level < len(_daughter_level):
            return _daughter_level[level]
        elif (level - 1) < len(_level_name_female):
            return "%s neta" % (_level_name_female[level - 1])
        else:
            return "%dª neta" % (level - 1)

    def get_aunt(self, level):
        if level < len(_sister_level):
            return _sister_level[level]
        elif (level - 2) < len(_level_name_female):
            return "%s tia avó" % (_level_name_female[level - 2])
        else:
            return "%dª tia avó" % (level - 2)

    def get_uncle(self, level):
        if level < len(_brother_level):
            return _brother_level[level]
        elif (level - 2) < len(_level_name_male_a):
            return "%s tio avô" % (_level_name_male_a[level - 2])
        else:
            return "%dº tio avô" % (level - 2)

    def get_nephew(self, level):
        if level < len(_nephew_level):
            return _nephew_level[level]
        elif (level - 1) < len(_level_name_male_a):
            return "%s sobrinho neto" % (_level_name_male_a[level - 1])
        else:
            return "%dº sobrinho neto" % (level - 1)

    def get_niece(self, level):
        if level < len(_niece_level):
            return _niece_level[level]
        elif (level - 1) < len(_level_name_female):
            return "%s sobrinha neta" % (_level_name_female[level - 1])
        else:
            return "%dª sobrinha neta" % (level - 1)

    def get_relationship(
        self, secondRel, firstRel, orig_person_gender, other_person_gender
    ):
        """
        returns a string representing the relationship between the two people,
        along with a list of common ancestors (typically father, mother)
        """

        common = ""
        if firstRel == 0:
            if secondRel == 0:
                return ("", common)
            elif other_person_gender == Person.MALE:
                return (self.get_father(secondRel), common)
            else:
                return (self.get_mother(secondRel), common)
        elif secondRel == 0:
            if other_person_gender == Person.MALE:
                return (self.get_son(firstRel), common)
            else:
                return (self.get_daughter(firstRel), common)
        elif firstRel == 1:
            if other_person_gender == Person.MALE:
                return (self.get_uncle(secondRel), common)
            else:
                return (self.get_aunt(secondRel), common)
        elif secondRel == 1:
            if other_person_gender == Person.MALE:
                return (self.get_nephew(firstRel - 1), common)
            else:
                return (self.get_niece(firstRel - 1), common)
        elif firstRel == secondRel == 2:
            if other_person_gender == Person.MALE:
                return ("primo irmão", common)
            else:
                return ("prima irmã", common)
        elif firstRel == secondRel:
            if other_person_gender == Person.MALE:
                return (self.get_male_cousin(firstRel - 1), common)
            else:
                return (self.get_female_cousin(firstRel - 1), common)
        elif firstRel == secondRel + 1:
            if other_person_gender == Person.MALE:
                return (self.get_distant_nephew(secondRel), common)
            else:
                return (self.get_distant_niece(secondRel), common)
        elif firstRel + 1 == secondRel:
            if other_person_gender == Person.MALE:
                return (self.get_distant_uncle(firstRel), common)
            else:
                return (self.get_distant_aunt(firstRel), common)
        else:
            if other_person_gender == Person.MALE:
                return (self.get_male_relative(firstRel, secondRel), common)
            else:
                return (self.get_female_relative(firstRel, secondRel), common)

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self.get_relationship(Ga, Gb, gender_a, gender_b)[0]

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        return self.get_relationship(1, 1, gender_a, gender_b)[0]

    # Relatório de parentesco

    def get_plural_relationship_string(
        self,
        Ga,
        Gb,
        reltocommon_a="",
        reltocommon_b="",
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        """
        Cria o objecto KinshipReport que produz o relatório.
        Os argumentos são:
        database        - a instância do banco de dados GRAMPS
        options_class   - instância da classe das opções para este relatório
        O presente relatório tem os seguintes parâmetros (variáveis de classe)
        que entram na classe de opções.
        maxdescend    - Máximo gerações de descendentes para incluir.
        maxascend     - Máximo de gerações ancestrais para incluir.
        incspouses    - Se deseja incluir cônjuges.
        inccousins    - Se deseja incluir primos.
        incaunts      - Se deseja incluir tios / sobrinhos.
        pid           - A identificação Gramps da pessoa central para o relatório.

        Preenche um mapa das matrizes contendo os descendentes
        da pessoa falecida. Esta função chama a si mesma recursivamente até
        atingir max_descend.
        Parâmetros:

        :param person_handle: o identificador da próxima pessoa
        :param Ga: O número de gerações, desde a pessoa principal até o
                   ancestral comum. É incrementado quando subir as gerações, e
                   deixado inalterado quando descer as gerações.
        :param Gb: O número de gerações desta pessoa (person_handle) até o
                   ancestral comum. É incrementado quando descer as
                   gerações and posto a zero quando subir as gerações.
        :param skip_handle: Identificador opcional para pular quando descer.
                            Isso é útil para pular o descendente que trouxe
                            essa generação em primeiro lugar.

        Preenche um mapa das matrizes contendo os ancestrais
        da pessoa falecida. Esta função chama a si mesma recursivamente até
        atingir max_ascend.
        Parâmetros:

        :param person_handle: o identificador da próxima pessoa
        :param Ga: O número de gerações, desde a pessoa principal até o
                   ancestral comum. É incrementado quando subir as gerações, e
                   deixado inalterado quando descer as gerações.
        :param Gb: O número de gerações desta pessoa (person_handle) até o
                   ancestral comum. É incrementado quando descer as
                   gerações and posto a zero quando subir as gerações.
        """

        rel_str = "???"

        if Ga == 0:
            # These are descendants

            if Gb < len(_CHILDREN_LEVEL):
                rel_str = _CHILDREN_LEVEL[Gb]
            else:
                rel_str = "descendentes"
        elif Gb == 0:
            # These are parents/grand parents

            if Ga < len(_PARENTS_LEVEL):
                rel_str = _PARENTS_LEVEL[Ga]
            else:
                rel_str = "ancestrais"
        elif Gb == 1:
            # These are siblings/aunts/uncles

            if Ga < len(_SIBLINGS_LEVEL):
                rel_str = _SIBLINGS_LEVEL[Ga]
            else:
                rel_str = "filhos dos ancestrais"
        elif Ga == 1:
            # These are nieces/nephews

            if Gb < len(_NEPHEWS_NIECES_LEVEL):
                rel_str = _NEPHEWS_NIECES_LEVEL[Gb - 1]
            else:
                rel_str = "sobrinhos sobrinhas"
        elif Ga > 1 and Ga == Gb:
            # These are cousins in the same generation

            if Ga == 2:
                rel_str = "primos e primas"
            elif Ga <= len(_level_name_plural):
                rel_str = "%s primos e primas" % _level_name_plural[Ga - 2]
            else:
                # security

                rel_str = "primos e primas"

        if in_law_b == True:
            rel_str = "cônjuges dos %s" % rel_str

        return rel_str


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_pt.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009-2010  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Place Tree View
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gui.views.listview import ListView, TEXT, ICON
from gramps.plugins.lib.libplaceview import PlaceBaseView
from gramps.gui.views.treemodels.placemodel import PlaceTreeModel
from gramps.gen.lib import Place, PlaceRef
from gramps.gen.errors import WindowActiveError
from gramps.gui.editors import EditPlace

# -------------------------------------------------------------------------
#
# Internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# PlaceTreeView
#
# -------------------------------------------------------------------------
class PlaceTreeView(PlaceBaseView):
    """
    A hierarchical view of the top three levels of places.
    """

    def __init__(self, pdata, dbstate, uistate):
        PlaceBaseView.__init__(
            self,
            pdata,
            dbstate,
            uistate,
            _("Place Tree View"),
            PlaceTreeModel,
            nav_group=0,
        )

    def get_viewtype_stock(self):
        """
        Override the default icon.  Set for hierarchical view.
        """
        return "gramps-tree-group"

    def define_actions(self):
        """
        Define actions for the popup menu specific to the tree view.
        """
        PlaceBaseView.define_actions(self)

        self._add_action("OpenBranch", self.open_branch)
        self._add_action("CloseBranch", self.close_branch)
        self._add_action("OpenAllNodes", self.open_all_nodes)
        self._add_action("CloseAllNodes", self.close_all_nodes)

    additional_ui = PlaceBaseView.additional_ui[:]
    additional_ui.append(
        """
        <section id="PopUpTree">
          <item>
            <attribute name="action">win.OpenBranch</attribute>
            <attribute name="label" translatable="yes">"""
        """Expand this Entire Group</attribute>
          </item>
          <item>
            <attribute name="action">win.CloseBranch</attribute>
            <attribute name="label" translatable="yes">"""
        """Collapse this Entire Group</attribute>
          </item>
          <item>
            <attribute name="action">win.OpenAllNodes</attribute>
            <attribute name="label" translatable="yes">"""
        """Expand all Nodes</attribute>
          </item>
          <item>
            <attribute name="action">win.CloseAllNodes</attribute>
            <attribute name="label" translatable="yes">"""
        """Collapse all Nodes</attribute>
          </item>
        </section>
        """
    )

    def add(self, *obj):
        """
        Add a new place.  Use the currently selected rows as parent places.
        """
        parent_list = []
        for handle in self.selected_handles():
            placeref = PlaceRef()
            placeref.ref = handle
            parent_list.append(placeref)

        place = Place()
        if len(parent_list) > 0:
            place.placeref_list = parent_list

        try:
            EditPlace(self.dbstate, self.uistate, [], place)
        except WindowActiveError:
            pass

    def row_update(self, handle_list):
        """
        Called when a place is updated.
        """
        ListView.row_update(self, handle_list)

        for handle in handle_list:
            # Rebuild the model if the primary parent has changed.
            if self._significant_change(handle):
                self.build_tree()
                break

    def _significant_change(self, handle):
        """
        Return True if the primary parent is different from the parent
        displayed in the tree, or if there are children.
        The first occurs if a change moves a child to a different parent,
        the second if a change to a parent occurs (a rename might shift
        position in the tree).
        """
        new_handle = None
        place = self.dbstate.db.get_place_from_handle(handle)
        placeref_list = place.get_placeref_list()
        if len(placeref_list) > 0:
            new_handle = placeref_list[0].ref

        old_handle = None
        children = False
        iter_ = self.model.get_iter_from_handle(handle)
        if iter_:
            parent_iter = self.model.iter_parent(iter_)
            if parent_iter:
                old_handle = self.model.get_handle_from_iter(parent_iter)
            children = self.model.get_node_from_iter(iter_).children
        return new_handle != old_handle or children

    def get_config_name(self):
        return __name__



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008       Brian G. Matherly
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This module provides the :class:`.Plugin` class for export plugins.
"""

from . import Plugin


class ExportPlugin(Plugin):
    """
    This class represents a plugin for exporting data from Gramps
    """

    def __init__(self, name, description, export_function, extension, config=None):
        """
        :param name: A friendly name to call this plugin.
            Example: "GEDCOM Export"
        :type name: string
        :param description: A short description of the plugin.
            Example: "This plugin will export a GEDCOM file from database"
        :type description: string
        :param export_function: A function to call to perform the export.
            The function must take the form:
                def export_function(database, filename, user, option_box):
            where:
                "db" is a Gramps database to import the data into
                "filename" is the file that the data will be exported to
                "user" provides UI output (callback, errors, etc)
        :type export_function: callable
        :param extension: The extension for the output file.
            Example: "ged"
        :type extension: str
        :param config: Options for the exporter
        :type config: tuple (??,??)
        :return: nothing
        """
        Plugin.__init__(self, name, description, export_function.__module__)
        self.__export_func = export_function
        self.__extension = extension
        self.__config = config

    def get_export_function(self):
        """
        Get the export function for this plugin.

        :return: the callable export_function passed into :meth:`__init__`
        """
        return self.__export_func

    def get_extension(self):
        """
        Get the file extension for the export file.

        :return: str
        """
        return self.__extension

    def get_config(self):
        """
        Get the config.

        :return: (??,??)
        """
        return self.__config



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2008       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Option class representing a document style.
"""

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from . import EnumeratedListOption
from ..docgen import StyleSheetList


# -------------------------------------------------------------------------
#
# StyleOption class
#
# -------------------------------------------------------------------------
class StyleOption(EnumeratedListOption):  # TODO this is likely dead code
    """
    This class describes an option that allows the use to select a style sheet.
    """

    def __init__(self, label, default_style, module_name):
        """
        :param label: A friendly label to be applied to this option.
            Example: "Style"
        :type label: string
        :param default_style: A docgen StyleSheet instance which provides the
        default styles.
        :type default_style: docgen StyleSheet
        :param module_name: The name of the module the style sheets belong to.
            Example: "web_cal"
        :type module_name: string
        :return: nothing
        """
        EnumeratedListOption.__init__(self, label, "default")

        self.__default_style = default_style
        self.__default_style.set_name("default")
        self.__style_file = "%s_style.xml" % module_name
        style_list = StyleSheetList(self.__style_file, self.__default_style)
        for style_name in style_list.get_style_names():
            self.add_item(style_name, style_name)

    def get_default_style(self):
        """Get the default style"""
        return self.__default_style

    def get_style_file(self):
        """Get the name of the style file"""
        return self.__style_file

    def get_style(self):
        """Get the selected style"""
        style_list = StyleSheetList(self.__style_file, self.__default_style)
        return style_list.get_style_sheet(self.get_value())
